@@ -25,6 +25,10 @@
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
  */
 
+#define LINE_ATTRIBUTES (GDK_GC_LINE_WIDTH|GDK_GC_LINE_STYLE| \
+			 GDK_GC_CAP_STYLE|GDK_GC_JOIN_STYLE)
+
+#include <config.h>
 #include <string.h>
 
 #include "gdkgc.h"
@@ -96,17 +100,17 @@ gdk_gc_win32_finalize (GObject *object)
 {
   GdkGCWin32 *win32_gc = GDK_GC_WIN32 (object);
   
-  if (win32_gc->clip_region)
-    gdk_region_destroy (win32_gc->clip_region);
+  if (win32_gc->hcliprgn != NULL)
+    DeleteObject (win32_gc->hcliprgn);
   
   if (win32_gc->values_mask & GDK_GC_FONT)
     gdk_font_unref (win32_gc->font);
   
   if (win32_gc->values_mask & GDK_GC_TILE)
-    gdk_drawable_unref (win32_gc->tile);
+    g_object_unref (win32_gc->tile);
   
   if (win32_gc->values_mask & GDK_GC_STIPPLE)
-    gdk_drawable_unref (win32_gc->stipple);
+    g_object_unref (win32_gc->stipple);
 
   if (win32_gc->pen_dashes)
     g_free (win32_gc->pen_dashes);
@@ -148,14 +152,10 @@ gdk_win32_gc_values_to_win32values (GdkGCValues    *values,
       win32_gc->font = values->font;
       if (win32_gc->font != NULL)
 	{
-	  gchar *xlfd;
-
 	  gdk_font_ref (win32_gc->font);
 	  win32_gc->values_mask |= GDK_GC_FONT;
-	  GDK_NOTE (GC, (xlfd = gdk_font_full_name_get (win32_gc->font),
-			 g_print ("%sfont=%s", s, xlfd),
-			 s = ",",
-			 gdk_font_full_name_free (xlfd)));
+	  GDK_NOTE (GC, (g_print ("%sfont=%p", s, win32_gc->font),
+			 s = ","));
 	}
       else
 	{
@@ -197,18 +197,19 @@ gdk_win32_gc_values_to_win32values (GdkGCValues    *values,
     {
       win32_gc->fill_style = values->fill;
       win32_gc->values_mask |= GDK_GC_FILL;
-      GDK_NOTE (GC, (g_print ("%sfill=%d", s, win32_gc->fill_style),
+      GDK_NOTE (GC, (g_print ("%sfill=%s", s,
+			      _gdk_win32_fill_style_to_string (win32_gc->fill_style)),
 		     s = ","));
     }
 
   if (mask & GDK_GC_TILE)
     {
       if (win32_gc->tile != NULL)
-	gdk_drawable_unref (win32_gc->tile);
+	g_object_unref (win32_gc->tile);
       win32_gc->tile = values->tile;
       if (win32_gc->tile != NULL)
 	{
-	  gdk_drawable_ref (win32_gc->tile);
+	  g_object_ref (win32_gc->tile);
 	  win32_gc->values_mask |= GDK_GC_TILE;
 	  GDK_NOTE (GC,
 		    (g_print ("%stile=%p", s,
@@ -226,7 +227,7 @@ gdk_win32_gc_values_to_win32values (GdkGCValues    *values,
   if (mask & GDK_GC_STIPPLE)
     {
       if (win32_gc->stipple != NULL)
-	gdk_drawable_unref (win32_gc->stipple);
+	g_object_unref (win32_gc->stipple);
       win32_gc->stipple = values->stipple;
       if (win32_gc->stipple != NULL)
 	{
@@ -265,7 +266,7 @@ gdk_win32_gc_values_to_win32values (GdkGCValues    *values,
 	    }
 	  else
 #endif
-	    gdk_drawable_ref (win32_gc->stipple);
+	    g_object_ref (win32_gc->stipple);
 	  win32_gc->values_mask |= GDK_GC_STIPPLE;
 	  GDK_NOTE (GC,
 		    (g_print ("%sstipple=%p", s,
@@ -282,22 +283,12 @@ gdk_win32_gc_values_to_win32values (GdkGCValues    *values,
 
   if (mask & GDK_GC_CLIP_MASK)
     {
-      if (win32_gc->clip_region != NULL)
-	{
-	  gdk_region_destroy (win32_gc->clip_region);
-	  win32_gc->clip_region = NULL;
-	}
-
       if (win32_gc->hcliprgn != NULL)
-	{
-	  DeleteObject (win32_gc->hcliprgn);
-	  win32_gc->hcliprgn = NULL;
-	}
+	DeleteObject (win32_gc->hcliprgn);
 
       if (values->clip_mask != NULL)
 	{
-	  win32_gc->hcliprgn =
-	    _gdk_win32_bitmap_to_region (values->clip_mask);
+	  win32_gc->hcliprgn = _gdk_win32_bitmap_to_hrgn (values->clip_mask);
 	  win32_gc->values_mask |= GDK_GC_CLIP_MASK;
 	}
       else
@@ -370,11 +361,9 @@ gdk_win32_gc_values_to_win32values (GdkGCValues    *values,
 	    {
 	      g_free (win32_gc->pen_dashes);
 	      win32_gc->pen_dashes = NULL;
-            win32_gc->pen_num_dashes = 0;
+	      win32_gc->pen_num_dashes = 0;
 	    }
           win32_gc->pen_style &= ~(PS_STYLE_MASK);
-	  GDK_NOTE (GC, (g_print ("%sps|=LINE_SOLID", s),
-			 s = ","));
 	  win32_gc->pen_style |= PS_SOLID;
 	  break;
 	case GDK_LINE_ON_OFF_DASH:
@@ -386,12 +375,12 @@ gdk_win32_gc_values_to_win32values (GdkGCValues    *values,
              * line endings ? --hb
              */
             win32_gc->pen_style &= ~(PS_STYLE_MASK);
-    	      GDK_NOTE (GC, (g_print ("%sps|=DASH", s),
-                                    s = ","));
-    	      win32_gc->pen_style |= PS_DASH;
+	    win32_gc->pen_style |= PS_DASH;
           }
 	  break;
 	}
+      GDK_NOTE (GC, (g_print ("%sps|=PS_STYLE_%s", s, _gdk_win32_psstyle_to_string (win32_gc->pen_style)),
+		     s = ","));
       win32_gc->values_mask |= GDK_GC_LINE_STYLE;
     }
 
@@ -402,21 +391,17 @@ gdk_win32_gc_values_to_win32values (GdkGCValues    *values,
 	{
 	case GDK_CAP_NOT_LAST:	/* ??? */
 	case GDK_CAP_BUTT:
-	  GDK_NOTE (GC, (g_print ("%sps|=ENDCAP_FLAT", s),
-			 s = ","));
 	  win32_gc->pen_style |= PS_ENDCAP_FLAT;
 	  break;
 	case GDK_CAP_ROUND:
-	  GDK_NOTE (GC, (g_print ("%sps|=ENDCAP_ROUND", s),
-			 s = ","));
 	  win32_gc->pen_style |= PS_ENDCAP_ROUND;
 	  break;
 	case GDK_CAP_PROJECTING:
-	  GDK_NOTE (GC, (g_print ("%sps|=ENDCAP_SQUARE", s),
-			 s = ","));
 	  win32_gc->pen_style |= PS_ENDCAP_SQUARE;
 	  break;
 	}
+      GDK_NOTE (GC, (g_print ("%sps|=PS_ENDCAP_%s", s, _gdk_win32_psendcap_to_string (win32_gc->pen_style)),
+		     s = ","));
       win32_gc->values_mask |= GDK_GC_CAP_STYLE;
     }
 
@@ -426,24 +411,20 @@ gdk_win32_gc_values_to_win32values (GdkGCValues    *values,
       switch (values->join_style)
 	{
 	case GDK_JOIN_MITER:
-	  GDK_NOTE (GC, (g_print ("%sps|=JOIN_MITER", s),
-			 s = ","));
 	  win32_gc->pen_style |= PS_JOIN_MITER;
 	  break;
 	case GDK_JOIN_ROUND:
-	  GDK_NOTE (GC, (g_print ("%sps|=JOIN_ROUND", s),
-			 s = ","));
 	  win32_gc->pen_style |= PS_JOIN_ROUND;
 	  break;
 	case GDK_JOIN_BEVEL:
-	  GDK_NOTE (GC, (g_print ("%sps|=JOIN_BEVEL", s),
-			 s = ","));
 	  win32_gc->pen_style |= PS_JOIN_BEVEL;
 	  break;
 	}
+      GDK_NOTE (GC, (g_print ("%sps|=PS_JOIN_%s", s, _gdk_win32_psjoin_to_string (win32_gc->pen_style)),
+		     s = ","));
       win32_gc->values_mask |= GDK_GC_JOIN_STYLE;
     }
-  GDK_NOTE (GC, g_print ("}\n"));
+  GDK_NOTE (GC, g_print ("} mask=(%s)", _gdk_win32_gcvalues_mask_to_string (win32_gc->values_mask)));
 }
 
 GdkGC*
@@ -462,8 +443,6 @@ _gdk_win32_gc_new (GdkDrawable	  *drawable,
   gc = g_object_new (_gdk_gc_win32_get_type (), NULL);
   win32_gc = GDK_GC_WIN32 (gc);
 
-  win32_gc->hdc = NULL;
-  win32_gc->clip_region = NULL;
   win32_gc->hcliprgn = NULL;
 
   /* Use the same default values as X11 does, even if they don't make
@@ -476,20 +455,22 @@ _gdk_win32_gc_new (GdkDrawable	  *drawable,
   win32_gc->fill_style = GDK_SOLID;
   win32_gc->tile = NULL;
   win32_gc->stipple = NULL;
-  win32_gc->pen_style = PS_GEOMETRIC|PS_ENDCAP_FLAT|PS_JOIN_MITER;
+  win32_gc->subwindow_mode = GDK_CLIP_BY_CHILDREN;
+  win32_gc->graphics_exposures = TRUE;
   win32_gc->pen_width = 0;
+  win32_gc->pen_style = PS_GEOMETRIC|PS_ENDCAP_FLAT|PS_JOIN_MITER;
   win32_gc->pen_dashes = NULL;
   win32_gc->pen_num_dashes = 0;
 
   win32_gc->values_mask = GDK_GC_FUNCTION | GDK_GC_FILL;
 
-  GDK_NOTE (GC, g_print ("_gdk_win32_gc_new: "));
+  GDK_NOTE (GC, g_print ("_gdk_win32_gc_new: %p: ", win32_gc));
   gdk_win32_gc_values_to_win32values (values, mask, win32_gc);
+  GDK_NOTE (GC, g_print ("\n"));
 
+  win32_gc->hdc = NULL;
   win32_gc->hwnd = NULL;
 
-  GDK_NOTE (GC, g_print (" = %p\n", gc));
-
   return gc;
 }
 
@@ -586,9 +567,9 @@ gdk_win32_gc_set_values (GdkGC           *gc,
 {
   g_return_if_fail (GDK_IS_GC (gc));
 
-  GDK_NOTE (GC, g_print ("gdk_win32_gc_set_values: "));
-
+  GDK_NOTE (GC, g_print ("gdk_win32_gc_set_values: %p: ", GDK_GC_WIN32 (gc)));
   gdk_win32_gc_values_to_win32values (values, mask, GDK_GC_WIN32 (gc));
+  GDK_NOTE (GC, g_print ("\n"));
 }
 
 static void
@@ -612,6 +593,8 @@ gdk_win32_gc_set_dashes (GdkGC *gc,
 
   win32_gc->pen_style |= (PS_GEOMETRIC | PS_USERSTYLE);
   win32_gc->pen_num_dashes = n;
+  if (win32_gc->pen_dashes != NULL)
+    g_free (win32_gc->pen_dashes);
   win32_gc->pen_dashes = g_new (DWORD, n);
   for (i = 0; i < n; i++)
     win32_gc->pen_dashes[i] = dash_list[i];
@@ -628,36 +611,35 @@ gdk_gc_set_clip_rectangle (GdkGC	*gc,
   win32_gc = GDK_GC_WIN32 (gc);
 
   if (win32_gc->hcliprgn)
-    {
-      DeleteObject (win32_gc->hcliprgn);
-      win32_gc->hcliprgn = NULL;
-    }
-
-  if (win32_gc->clip_region)
-    gdk_region_destroy (win32_gc->clip_region);
+    DeleteObject (win32_gc->hcliprgn);
 
   if (rectangle)
     {
-      GDK_NOTE (GC,
-		g_print ("gdk_gc_set_clip_rectangle: %dx%d@+%d+%d\n",
-			 rectangle->width, rectangle->height,
-			 rectangle->x, rectangle->y));
-      win32_gc->clip_region = gdk_region_rectangle (rectangle);
+      GDK_NOTE (GC, g_print ("gdk_gc_set_clip_rectangle: %p: %s\n",
+			     win32_gc,
+			     _gdk_win32_gdkrectangle_to_string (rectangle)));
+      win32_gc->hcliprgn = CreateRectRgn (rectangle->x, rectangle->y,
+					  rectangle->x + rectangle->width,
+					  rectangle->y + rectangle->height);
       win32_gc->values_mask |= GDK_GC_CLIP_MASK;
     }
   else
     {
       GDK_NOTE (GC, g_print ("gdk_gc_set_clip_rectangle: NULL\n"));
 
-      win32_gc->clip_region = NULL;
+      win32_gc->hcliprgn = NULL;
       win32_gc->values_mask &= ~GDK_GC_CLIP_MASK;
     }
-    win32_gc->values_mask &= ~(GDK_GC_CLIP_X_ORIGIN | GDK_GC_CLIP_Y_ORIGIN);
+
+  gc->clip_x_origin = 0;
+  gc->clip_y_origin = 0;
+  
+  win32_gc->values_mask &= ~(GDK_GC_CLIP_X_ORIGIN | GDK_GC_CLIP_Y_ORIGIN);
 } 
 
 void
-gdk_gc_set_clip_region (GdkGC		 *gc,
-			GdkRegion	 *region)
+gdk_gc_set_clip_region (GdkGC	  *gc,
+			GdkRegion *region)
 {
   GdkGCWin32 *win32_gc;
 
@@ -665,30 +647,29 @@ gdk_gc_set_clip_region (GdkGC		 *gc,
 
   win32_gc = GDK_GC_WIN32 (gc);
 
-  if (win32_gc->clip_region)
-    gdk_region_destroy (win32_gc->clip_region);
+  if (win32_gc->hcliprgn)
+    DeleteObject (win32_gc->hcliprgn);
 
   if (region)
     {
-      GDK_NOTE (GC, g_print ("gdk_gc_set_clip_region: %dx%d+%d+%d\n",
-			       region->extents.x2 - region->extents.x1,
-			       region->extents.y2 - region->extents.y1,
-			       region->extents.x1, region->extents.y1));
+      GDK_NOTE (GC, g_print ("gdk_gc_set_clip_region: %p: %s\n",
+			     win32_gc,
+			     _gdk_win32_gdkregion_to_string (region)));
 
-      win32_gc->clip_region = gdk_region_copy (region);
+      win32_gc->hcliprgn = _gdk_win32_gdkregion_to_hrgn (region, 0, 0);
       win32_gc->values_mask |= GDK_GC_CLIP_MASK;
     }
   else
     {
       GDK_NOTE (GC, g_print ("gdk_gc_set_clip_region: NULL\n"));
 
-      win32_gc->clip_region = NULL;
+      win32_gc->hcliprgn = NULL;
       win32_gc->values_mask &= ~GDK_GC_CLIP_MASK;
     }
 
   gc->clip_x_origin = 0;
   gc->clip_y_origin = 0;
-
+  
   win32_gc->values_mask &= ~(GDK_GC_CLIP_X_ORIGIN | GDK_GC_CLIP_Y_ORIGIN);
 }
 
@@ -705,24 +686,35 @@ gdk_gc_copy (GdkGC *dst_gc,
   dst_win32_gc = GDK_GC_WIN32 (dst_gc);
   src_win32_gc = GDK_GC_WIN32 (src_gc);
 
-  if (dst_win32_gc->font != NULL)
-    gdk_font_unref (dst_win32_gc->font);
-  if (dst_win32_gc->tile != NULL)
-    gdk_drawable_unref (dst_win32_gc->tile);
-  if (dst_win32_gc->stipple != NULL)
-    gdk_drawable_unref (dst_win32_gc->stipple);
-  if (dst_win32_gc->clip_region != NULL)
-    gdk_region_destroy (dst_win32_gc->clip_region);
+  GDK_NOTE (GC, g_print ("gdk_gc_copy: %p := %p\n", dst_win32_gc, src_win32_gc));
+
+  if (dst_gc->colormap)
+    g_object_unref (G_OBJECT (dst_gc->colormap));
+
   if (dst_win32_gc->hcliprgn != NULL)
     DeleteObject (dst_win32_gc->hcliprgn);
+
+  if (dst_win32_gc->font != NULL)
+    gdk_font_unref (dst_win32_gc->font);
+
+  if (dst_win32_gc->tile != NULL)
+    g_object_unref (dst_win32_gc->tile);
+
+  if (dst_win32_gc->stipple != NULL)
+    g_object_unref (dst_win32_gc->stipple);
+
   if (dst_win32_gc->pen_dashes)
     g_free (dst_win32_gc->pen_dashes);
   
-  *dst_win32_gc = *src_win32_gc;
+  dst_gc->clip_x_origin = src_gc->clip_x_origin;
+  dst_gc->clip_y_origin = src_gc->clip_y_origin;
+  dst_gc->ts_x_origin = src_gc->ts_x_origin;
+  dst_gc->ts_y_origin = src_gc->ts_y_origin;
+  dst_gc->colormap = src_gc->colormap;
+  if (dst_gc->colormap)
+    g_object_ref (G_OBJECT (dst_gc->colormap));
 
-  if (dst_win32_gc->pen_dashes)
-    dst_win32_gc->pen_dashes = g_memdup (src_win32_gc->pen_dashes, 
-                                         sizeof (DWORD) * src_win32_gc->pen_num_dashes);
+  dst_win32_gc->hcliprgn = src_win32_gc->hcliprgn;
   if (dst_win32_gc->hcliprgn)
     {
       /* create a new region, to copy to */
@@ -731,14 +723,47 @@ gdk_gc_copy (GdkGC *dst_gc,
       CombineRgn (dst_win32_gc->hcliprgn, src_win32_gc->hcliprgn,
 		  NULL, RGN_COPY);
     }
-  if (dst_win32_gc->clip_region != NULL)
-    dst_win32_gc->clip_region = gdk_region_copy (dst_win32_gc->clip_region);
+
+  dst_win32_gc->values_mask = src_win32_gc->values_mask; 
+  dst_win32_gc->foreground = src_win32_gc->foreground;
+  dst_win32_gc->background = src_win32_gc->background;
+  dst_win32_gc->font = src_win32_gc->font;
   if (dst_win32_gc->font != NULL)
     gdk_font_ref (dst_win32_gc->font);
+
+  dst_win32_gc->rop2 = src_win32_gc->rop2;
+  dst_win32_gc->fill_style = src_win32_gc->fill_style;
+  dst_win32_gc->tile = src_win32_gc->tile;
   if (dst_win32_gc->tile != NULL)
-    gdk_drawable_ref (dst_win32_gc->tile);
+    g_object_ref (dst_win32_gc->tile);
+
+  dst_win32_gc->stipple = src_win32_gc->stipple;
   if (dst_win32_gc->stipple != NULL)
-    gdk_drawable_ref (dst_win32_gc->stipple);
+    g_object_ref (dst_win32_gc->stipple);
+
+  dst_win32_gc->subwindow_mode = src_win32_gc->subwindow_mode;
+  dst_win32_gc->graphics_exposures = src_win32_gc->graphics_exposures;
+  dst_win32_gc->pen_width = src_win32_gc->pen_width;
+  dst_win32_gc->pen_style = src_win32_gc->pen_style;
+  dst_win32_gc->pen_dashes = src_win32_gc->pen_dashes;
+  if (dst_win32_gc->pen_dashes)
+    dst_win32_gc->pen_dashes = g_memdup (src_win32_gc->pen_dashes, 
+                                         sizeof (DWORD) * src_win32_gc->pen_num_dashes);
+  dst_win32_gc->pen_num_dashes = src_win32_gc->pen_num_dashes;
+
+
+  dst_win32_gc->hdc = NULL;
+  dst_win32_gc->saved_dc = FALSE;
+  dst_win32_gc->hwnd = NULL;
+  dst_win32_gc->holdpal = NULL;
+}
+
+GdkScreen *  
+gdk_gc_get_screen (GdkGC *gc)
+{
+  g_return_val_if_fail (GDK_IS_GC_WIN32 (gc), NULL);
+  
+  return _gdk_screen;
 }
 
 static guint bitmask[9] = { 0, 1, 3, 7, 15, 31, 63, 127, 255 };
@@ -781,15 +806,12 @@ _gdk_win32_colormap_color (GdkColormap *colormap,
     }
 }
 
-static void
+static COLORREF
 predraw_set_foreground (GdkGC       *gc,
 			GdkColormap *colormap,
 			gboolean    *ok)
 {
   COLORREF fg;
-  LOGBRUSH logbrush;
-  HPEN hpen;
-  HBRUSH hbr;
   GdkGCWin32 *win32_gc = (GdkGCWin32 *) gc;
   GdkColormapPrivateWin32 *colormap_private;
   gint k;
@@ -814,88 +836,59 @@ predraw_set_foreground (GdkGC       *gc,
   fg = _gdk_win32_colormap_color (colormap, win32_gc->foreground);
 
   GDK_NOTE (GC, g_print ("predraw_set_foreground: fg=%06lx\n", fg));
-
-  if (SetTextColor (win32_gc->hdc, fg) == CLR_INVALID)
-    WIN32_GDI_FAILED ("SetTextColor"), *ok = FALSE;
-
-  /* Create and select pen and brush. */
-
-  logbrush.lbStyle = BS_SOLID;
-  logbrush.lbColor = fg;
-  logbrush.lbHatch = 0;
-
-  if (win32_gc->pen_num_dashes > 0 && !IS_WIN_NT ())
-    {
-      /* The Win9x GDI is rather limited so we either draw dotted
-       * lines ourselve (only horizontal and vertical) or let them
-       * be drawn solid to avoid implementing a whole line renderer
-       */
-      if (*ok && (hpen = ExtCreatePen (
-                           (win32_gc->pen_style & ~(PS_STYLE_MASK)) | PS_SOLID,
-                           (win32_gc->pen_width > 0 ? win32_gc->pen_width : 1),
-                           &logbrush, 
-                           0, NULL)) == NULL)
-        WIN32_GDI_FAILED ("ExtCreatePen"), *ok = FALSE;
-    }
-  else
-    {
-      if (*ok && (hpen = ExtCreatePen (win32_gc->pen_style,
-                           (win32_gc->pen_width > 0 ? win32_gc->pen_width : 1),
-                           &logbrush, 
-                           win32_gc->pen_num_dashes,
-                           win32_gc->pen_dashes)) == NULL)
-        WIN32_GDI_FAILED ("ExtCreatePen"), *ok = FALSE;
-    }
-  
-  if (*ok && SelectObject (win32_gc->hdc, hpen) == NULL)
-    WIN32_GDI_FAILED ("SelectObject"), *ok = FALSE;
-
-  switch (win32_gc->fill_style)
-    {
-    case GDK_OPAQUE_STIPPLED:
-      if (*ok && (hbr = CreatePatternBrush (GDK_PIXMAP_HBITMAP (win32_gc->stipple))) == NULL)
-	WIN32_GDI_FAILED ("CreatePatternBrush"), *ok = FALSE;
-      if (*ok && !SetBrushOrgEx(win32_gc->hdc, gc->ts_x_origin,
-				gc->ts_y_origin, NULL))
-	WIN32_GDI_FAILED ("SetBrushOrgEx"), *ok = FALSE;
-	
-      break;
-
-    case GDK_SOLID:
-    default:
-      if (*ok && (hbr = CreateSolidBrush (fg)) == NULL)
-	WIN32_GDI_FAILED ("CreateSolidBrush"), *ok = FALSE;
-      break;
-  }
-  if (*ok)
-    {
-      HBRUSH old_hbr = SelectObject (win32_gc->hdc, hbr);
-      if (old_hbr == NULL)
-	WIN32_GDI_FAILED ("SelectObject"), *ok = FALSE;
-    }
-}  
-
-static void
-predraw_set_background (GdkGC       *gc,
-			GdkColormap *colormap,
-			gboolean    *ok)
-{
-  GdkGCWin32 *win32_gc = (GdkGCWin32 *) gc;
-
-  if (win32_gc->values_mask & GDK_GC_BACKGROUND)
-    {
-      COLORREF bg = _gdk_win32_colormap_color (colormap, win32_gc->background);
-
-      if (SetBkColor (win32_gc->hdc, bg) == CLR_INVALID)
-        WIN32_GDI_FAILED ("SetBkColor"), *ok = FALSE;
-    }
-  else
-    {
-      if (!SetBkMode (win32_gc->hdc, TRANSPARENT))
-        WIN32_GDI_FAILED ("SetBkMode"), *ok = FALSE;
-    }
+  return fg;
 }
 
+/**
+ * gdk_win32_hdc_get:
+ * @drawable: destination #GdkDrawable
+ * @gc: #GdkGC to use for drawing on @drawable
+ * @usage: mask indicating what properties needs to be set up
+ *
+ * Allocates a Windows device context handle (HDC) for drawing into
+ * @drawable, and sets it up appropriately according to @usage.
+ *
+ * Each #GdkGC can at one time have only one HDC associated with it.
+ *
+ * The following flags in @mask are handled:
+ *
+ * If %GDK_GC_FOREGROUND is set in @mask, a solid brush of the
+ * foreground color in @gc is selected into the HDC. The text color of
+ * the HDC is also set. If the @drawable has a palette (256-color
+ * mode), the palette is selected and realized.
+ *
+ * If any of the line attribute flags (%GDK_GC_LINE_WIDTH,
+ * %GDK_GC_LINE_STYLE, %GDK_GC_CAP_STYLE and %GDK_GC_JOIN_STYLE) is
+ * set in @mask, a solid pen of the foreground color and appropriate
+ * width and stule is created and selected into the HDC. Note that the
+ * dash properties are not completely implemented.
+ *
+ * If the %GDK_GC_FONT flag is set, the background mix mode is set to
+ * %TRANSPARENT. and the text alignment is set to
+ * %TA_BASELINE|%TA_LEFT. Note that no font gets selected into the HDC
+ * by this function.
+ *
+ * Some things are done regardless of @mask: If the function in @gc is
+ * any other than %GDK_COPY, the raster operation of the HDC is
+ * set. If @gc has a clip mask, the clip region of the HDC is set.
+ *
+ * Note that the fill style, tile, stipple, and tile and stipple
+ * origins in the @gc are ignored by this function. (In general, tiles
+ * and stipples can't be implemented directly on Win32; you need to do
+ * multiple pass drawing and blitting to implement tiles or
+ * stipples. GDK does just that when you call the GDK drawing
+ * functions with a GC that asks for tiles or stipples.)
+ *
+ * When the HDC is no longer used, it should be released by calling
+ * <function>gdk_win32_hdc_release()</function> with the same
+ * parameters.
+ *
+ * If you modify the HDC by calling <function>SelectObject</function>
+ * you should undo those modifications before calling
+ * <function>gdk_win32_hdc_release()</function>.
+ *
+ * Return value: The HDC.
+ **/
 HDC
 gdk_win32_hdc_get (GdkDrawable    *drawable,
 		   GdkGC          *gc,
@@ -904,7 +897,10 @@ gdk_win32_hdc_get (GdkDrawable    *drawable,
   GdkGCWin32 *win32_gc = (GdkGCWin32 *) gc;
   GdkDrawableImplWin32 *impl = NULL;
   gboolean ok = TRUE;
-  int flag;
+  COLORREF fg = RGB (0, 0, 0);
+  LOGBRUSH logbrush;
+  HPEN hpen;
+  HBRUSH hbr;
 
   g_assert (win32_gc->hdc == NULL);
 
@@ -940,185 +936,96 @@ gdk_win32_hdc_get (GdkDrawable    *drawable,
     }
   
   if (ok && (usage & GDK_GC_FOREGROUND))
-    predraw_set_foreground (gc, impl->colormap, &ok);
+    {
+      fg = predraw_set_foreground (gc, impl->colormap, &ok);
+      if (ok && (hbr = CreateSolidBrush (fg)) == NULL)
+	WIN32_GDI_FAILED ("CreateSolidBrush"), ok = FALSE;
+
+      if (ok && SelectObject (win32_gc->hdc, hbr) == NULL)
+	WIN32_GDI_FAILED ("SelectObject"), ok = FALSE;
+
+      if (ok && SetTextColor (win32_gc->hdc, fg) == CLR_INVALID)
+	WIN32_GDI_FAILED ("SetTextColor"), ok = FALSE;
+    }
+
+  if (ok && (usage & LINE_ATTRIBUTES))
+    {
+      /* Create and select pen */
+      logbrush.lbStyle = BS_SOLID;
+      logbrush.lbColor = fg;
+      logbrush.lbHatch = 0;
+      
+      if (win32_gc->pen_num_dashes > 0 && !IS_WIN_NT ())
+	{
+	  /* The Win9x GDI is rather limited so we either draw dashed
+	   * lines ourselves (only horizontal and vertical) or let them be
+	   * drawn solid to avoid implementing a whole line renderer.
+	   */
+	  if ((hpen = ExtCreatePen (
+				    (win32_gc->pen_style & ~(PS_STYLE_MASK)) | PS_SOLID,
+				    MAX (win32_gc->pen_width, 1),
+				    &logbrush, 
+				    0, NULL)) == NULL)
+	    WIN32_GDI_FAILED ("ExtCreatePen"), ok = FALSE;
+	}
+      else
+	{
+	  if ((hpen = ExtCreatePen (win32_gc->pen_style,
+				    MAX (win32_gc->pen_width, 1),
+				    &logbrush, 
+				    win32_gc->pen_num_dashes,
+				    win32_gc->pen_dashes)) == NULL)
+	    WIN32_GDI_FAILED ("ExtCreatePen"), ok = FALSE;
+	}
+      
+      if (ok && SelectObject (win32_gc->hdc, hpen) == NULL)
+	WIN32_GDI_FAILED ("SelectObject"), ok = FALSE;
+    }
 
-  if (ok && (usage & GDK_GC_BACKGROUND))
-    predraw_set_background (gc, impl->colormap, &ok);
-  
   if (ok && (usage & GDK_GC_FONT))
     {
       if (SetBkMode (win32_gc->hdc, TRANSPARENT) == 0)
 	WIN32_GDI_FAILED ("SetBkMode"), ok = FALSE;
   
-      if (ok && SetTextAlign (win32_gc->hdc, TA_BASELINE) == GDI_ERROR)
+      if (ok && SetTextAlign (win32_gc->hdc, TA_BASELINE|TA_LEFT|TA_NOUPDATECP) == GDI_ERROR)
 	WIN32_GDI_FAILED ("SetTextAlign"), ok = FALSE;
     }
   
-  if (ok && (win32_gc->values_mask & GDK_GC_FUNCTION))
+  if (ok && win32_gc->rop2 != R2_COPYPEN)
     if (SetROP2 (win32_gc->hdc, win32_gc->rop2) == 0)
       WIN32_GDI_FAILED ("SetROP2"), ok = FALSE;
 
-  if (win32_gc->values_mask & GDK_GC_CLIP_MASK)
-    g_assert ((win32_gc->clip_region != NULL) != (win32_gc->hcliprgn != NULL));
-
-  if (ok
-      && (win32_gc->values_mask & GDK_GC_CLIP_MASK)
-      && win32_gc->clip_region != NULL)
-    {
-      HRGN hrgn;
-      RGNDATA *rgndata;
-      RECT *rect;
-      GdkRegionBox *boxes = win32_gc->clip_region->rects;
-      guint nbytes =
-	sizeof (RGNDATAHEADER) + (sizeof (RECT) * win32_gc->clip_region->numRects);
-      int i;
-
-      rgndata =	g_malloc (nbytes);
-      rgndata->rdh.dwSize = sizeof (RGNDATAHEADER);
-      rgndata->rdh.iType = RDH_RECTANGLES;
-      rgndata->rdh.nCount = rgndata->rdh.nRgnSize = 0;
-      SetRect (&rgndata->rdh.rcBound,
-	       G_MAXSHORT, G_MAXSHORT, G_MINSHORT, G_MINSHORT);
-
-      for (i = 0; i < win32_gc->clip_region->numRects; i++)
-	{
-	  rect = ((RECT *) rgndata->Buffer) + rgndata->rdh.nCount++;
-
-	  rect->left = CLAMP (boxes[i].x1 + gc->clip_x_origin,
-			      G_MINSHORT, G_MAXSHORT);
-	  rect->right = CLAMP (boxes[i].x2 + gc->clip_x_origin,
-			       G_MINSHORT, G_MAXSHORT);
-	  rect->top = CLAMP (boxes[i].y1 + gc->clip_y_origin,
-			     G_MINSHORT, G_MAXSHORT);
-	  rect->bottom = CLAMP (boxes[i].y2 + gc->clip_y_origin,
-				G_MINSHORT, G_MAXSHORT);
-
-	  GDK_NOTE (GC, g_print ("clip rgn box %d: %ldx%ld@+%ld+%ld\n",
-				 i,
-				 rect->right-rect->left,
-				 rect->bottom-rect->top,
-				 rect->left, rect->top));
-
-	  if (rect->left < rgndata->rdh.rcBound.left)
-	    rgndata->rdh.rcBound.left = rect->left;
-	  if (rect->right > rgndata->rdh.rcBound.right)
-	    rgndata->rdh.rcBound.right = rect->right;
-	  if (rect->top < rgndata->rdh.rcBound.top)
-	    rgndata->rdh.rcBound.top = rect->top;
-	  if (rect->bottom > rgndata->rdh.rcBound.bottom)
-	    rgndata->rdh.rcBound.bottom = rect->bottom;
-	}
-      if ((hrgn = ExtCreateRegion (NULL, nbytes, rgndata)) == NULL)
-	WIN32_API_FAILED ("ExtCreateRegion"), ok = FALSE;
-
-      g_free (rgndata);
-
-      /*
-       * XXX: combine the fillmode-stipple with the clip region possibly 
-       *  this needs to be done with the hcliprgn below as well, but 
-       *  the test case here were check boxes and radio buttons ...
-       */
-      if (   (win32_gc->values_mask & GDK_GC_STIPPLE)
-          && (win32_gc->values_mask & GDK_GC_FILL) && (win32_gc->fill_style == GDK_STIPPLED))
-        {
-          HRGN hstipplergn = _gdk_win32_bitmap_to_region (win32_gc->stipple);
-
-          if (OffsetRgn (hstipplergn, 
-             win32_gc->values_mask & GDK_GC_TS_X_ORIGIN ? gc->ts_x_origin : 0,
-             win32_gc->values_mask & GDK_GC_TS_Y_ORIGIN ? gc->ts_y_origin : 0) == ERROR)
-            WIN32_API_FAILED ("OffsetRgn"), ok = FALSE;
-          if (CombineRgn (hrgn, hrgn, hstipplergn, RGN_AND) == ERROR)
-            WIN32_API_FAILED ("CombineRgn"), ok = FALSE;
-          if (!DeleteObject (hstipplergn))
-            WIN32_API_FAILED ("DeleteObject");
-        }
-      if (ok && SelectClipRgn (win32_gc->hdc, hrgn) == ERROR)
-	WIN32_API_FAILED ("SelectClipRgn"), ok = FALSE;
-
-      if (hrgn != NULL)
-	DeleteObject (hrgn);
-    }
-  else if (ok
-	   && (win32_gc->values_mask & GDK_GC_CLIP_MASK)
-	   && win32_gc->hcliprgn != NULL)
+  if (ok &&
+      (win32_gc->values_mask & GDK_GC_CLIP_MASK) &&
+      win32_gc->hcliprgn != NULL)
     {
       if (SelectClipRgn (win32_gc->hdc, win32_gc->hcliprgn) == ERROR)
 	WIN32_API_FAILED ("SelectClipRgn"), ok = FALSE;
-      if (ok && !OffsetClipRgn (win32_gc->hdc,
-		 win32_gc->values_mask & GDK_GC_CLIP_X_ORIGIN ? gc->clip_x_origin : 0,
-		 win32_gc->values_mask & GDK_GC_CLIP_Y_ORIGIN ? gc->clip_y_origin : 0))
+
+      if (ok && win32_gc->values_mask & (GDK_GC_CLIP_X_ORIGIN | GDK_GC_CLIP_Y_ORIGIN) &&
+	  OffsetClipRgn (win32_gc->hdc,
+	    win32_gc->values_mask & GDK_GC_CLIP_X_ORIGIN ? gc->clip_x_origin : 0,
+	    win32_gc->values_mask & GDK_GC_CLIP_Y_ORIGIN ? gc->clip_y_origin : 0) == ERROR)
 	WIN32_API_FAILED ("OffsetClipRgn"), ok = FALSE;
     }
-#ifdef G_ENABLE_DEBUG
-  if (_gdk_debug_flags & GDK_DEBUG_GC)
-    {
-      HGDIOBJ obj;
-      LOGBRUSH logbrush;
-      EXTLOGPEN extlogpen;
-      HRGN hrgn;
-      RECT rect;
 
-      g_print ("gdk_win32_hdc_get: %p\n", win32_gc->hdc);
-      obj = GetCurrentObject (win32_gc->hdc, OBJ_BRUSH);
-      GetObject (obj, sizeof (LOGBRUSH), &logbrush);
-      g_print ("brush: style: %s color: %06lx hatch: %#lx\n",
-	       (logbrush.lbStyle == BS_HOLLOW ? "HOLLOW" :
-		(logbrush.lbStyle == BS_PATTERN ? "PATTERN" :
-		 (logbrush.lbStyle == BS_SOLID ? "SOLID" :
-		  "???"))),
-	       logbrush.lbColor,
-	       logbrush.lbHatch);
-      obj = GetCurrentObject (win32_gc->hdc, OBJ_PEN);
-      GetObject (obj, sizeof (EXTLOGPEN), &extlogpen);
-      g_print ("pen: type: %s style: %s endcap: %s join: %s width: %d brush: %s\n",
-	       ((extlogpen.elpPenStyle & PS_TYPE_MASK) ==
-		PS_GEOMETRIC ? "GEOMETRIC" : "COSMETIC"),
-	       ((extlogpen.elpPenStyle & PS_STYLE_MASK) ==
-		PS_NULL ? "NULL" :
-		((extlogpen.elpPenStyle & PS_STYLE_MASK) == PS_SOLID ? "SOLID" :
-		 ((extlogpen.elpPenStyle & PS_STYLE_MASK) == PS_DOT ? "DOT" :
-		  ((extlogpen.elpPenStyle & PS_STYLE_MASK) == PS_DASH ? "DASH" :
- 		   "???")))),
-	       ((extlogpen.elpPenStyle & PS_ENDCAP_MASK) ==
-		PS_ENDCAP_FLAT ? "FLAT" :
-		((extlogpen.elpPenStyle & PS_ENDCAP_MASK) ==
-		 PS_ENDCAP_ROUND ? "ROUND" :
-		 ((extlogpen.elpPenStyle & PS_ENDCAP_MASK) ==
-		  PS_ENDCAP_SQUARE ? "ROUND" :
-		  ((extlogpen.elpPenStyle & PS_ENDCAP_MASK) ==
-		   PS_ENDCAP_SQUARE ? "ROUND" : "???")))),
-	       ((extlogpen.elpPenStyle & PS_JOIN_MASK) ==
-		PS_JOIN_BEVEL ? "BEVEL" :
-		((extlogpen.elpPenStyle & PS_JOIN_MASK) ==
-		 PS_JOIN_MITER ? "MITER" :
-		 ((extlogpen.elpPenStyle & PS_JOIN_MASK) ==
-		  PS_JOIN_ROUND ? "ROUND" : "???"))),
-	       extlogpen.elpWidth,
-	       (extlogpen.elpBrushStyle == BS_DIBPATTERN ? "DIBPATTERN" :
-		(extlogpen.elpBrushStyle == BS_DIBPATTERNPT ? "DIBPATTERNPT" :
-		 (extlogpen.elpBrushStyle == BS_HATCHED ? "HATCHED" :
-		  (extlogpen.elpBrushStyle == BS_HOLLOW ? "HOLLOW" :
-		   (extlogpen.elpBrushStyle == BS_PATTERN ? "PATTERN" :
-		    (extlogpen.elpBrushStyle == BS_SOLID ? "SOLID" : "???")))))));
-      hrgn = CreateRectRgn (0, 0, 0, 0);
-      if ((flag = GetClipRgn (win32_gc->hdc, hrgn)) == -1)
-	WIN32_API_FAILED ("GetClipRgn");
-      else if (flag == 0)
-	g_print ("no clip region\n");
-      else if (flag == 1)
-	{
-	  GetRgnBox (hrgn, &rect);
-	  g_print ("clip region bbox: %ldx%ld@+%ld+%ld\n",
-		   rect.right - rect.left,
-		   rect.bottom - rect.top,
-		   rect.left, rect.top);
-	}
-      DeleteObject (hrgn);
-    }
-#endif
+  GDK_NOTE (GC, (g_print ("gdk_win32_hdc_get: %p (%s): ",
+			  win32_gc, _gdk_win32_gcvalues_mask_to_string (usage)),
+		 _gdk_win32_print_dc (win32_gc->hdc)));
+
   return win32_gc->hdc;
 }
 
+/**
+ * gdk_win32_hdc_release:
+ * @drawable: destination #GdkDrawable
+ * @gc: #GdkGC to use for drawing on @drawable
+ * @usage: mask indicating what properties were set up
+ *
+ * This function deallocates the Windows device context allocated by
+ * <funcion>gdk_win32_hdc_get()</function>. It should be called with
+ * the same parameters.
+ **/
 void
 gdk_win32_hdc_release (GdkDrawable    *drawable,
 		       GdkGC          *gc,
@@ -1129,7 +1036,9 @@ gdk_win32_hdc_release (GdkDrawable    *drawable,
   HGDIOBJ hpen = NULL;
   HGDIOBJ hbr = NULL;
 
-  GDK_NOTE (GC, g_print ("gdk_win32_hdc_release: %p\n", win32_gc->hdc));
+  GDK_NOTE (GC, g_print ("gdk_win32_hdc_release: %p: %p (%s)\n",
+			 win32_gc, win32_gc->hdc,
+			 _gdk_win32_gcvalues_mask_to_string (usage)));
 
   if (GDK_IS_DRAWABLE_IMPL_WIN32 (drawable))
     impl = GDK_DRAWABLE_IMPL_WIN32(drawable);
@@ -1154,36 +1063,26 @@ gdk_win32_hdc_release (GdkDrawable    *drawable,
       win32_gc->holdpal = NULL;
     }
 
+  if (usage & LINE_ATTRIBUTES)
+    if ((hpen = GetCurrentObject (win32_gc->hdc, OBJ_PEN)) == NULL)
+      WIN32_GDI_FAILED ("GetCurrentObject");
+  
   if (usage & GDK_GC_FOREGROUND)
-    {
-      if ((hpen = GetCurrentObject (win32_gc->hdc, OBJ_PEN)) == NULL)
-	WIN32_GDI_FAILED ("GetCurrentObject");
+    if ((hbr = GetCurrentObject (win32_gc->hdc, OBJ_BRUSH)) == NULL)
+      WIN32_GDI_FAILED ("GetCurrentObject");
 
-      if ((hbr = GetCurrentObject (win32_gc->hdc, OBJ_BRUSH)) == NULL)
-	WIN32_GDI_FAILED ("GetCurrentObject");
-    }
-
-  if (!RestoreDC (win32_gc->hdc, win32_gc->saved_dc))
-    WIN32_GDI_FAILED ("RestoreDC");
+  GDI_CALL (RestoreDC, (win32_gc->hdc, win32_gc->saved_dc));
 
   if (GDK_IS_PIXMAP_IMPL_WIN32 (impl))
-    {
-      if (!DeleteDC (win32_gc->hdc))
-	WIN32_GDI_FAILED ("DeleteDC");
-    }
+    GDI_CALL (DeleteDC, (win32_gc->hdc));
   else
-    {
-      if (!ReleaseDC (win32_gc->hwnd, win32_gc->hdc))
-	WIN32_GDI_FAILED ("ReleaseDC");
-    }
+    GDI_CALL (ReleaseDC, (win32_gc->hwnd, win32_gc->hdc));
 
   if (hpen != NULL)
-    if (!DeleteObject (hpen))
-      WIN32_GDI_FAILED ("DeleteObject");
+    GDI_CALL (DeleteObject, (hpen));
   
   if (hbr != NULL)
-    if (!DeleteObject (hbr))
-      WIN32_GDI_FAILED ("DeleteObject");
+    GDI_CALL (DeleteObject, (hbr));
 
   win32_gc->hdc = NULL;
 }
@@ -1195,23 +1094,28 @@ gdk_win32_hdc_release (GdkDrawable    *drawable,
  * bitmaps (and not pixmaps), with zero pixels being transparent.
  */
 
-/* _gdk_win32_bitmap_to_region : Create a region from the
+/* _gdk_win32_bitmap_to_hrgn : Create a region from the
  * "non-transparent" pixels of a bitmap.
  */
 
 HRGN
-_gdk_win32_bitmap_to_region (GdkPixmap *pixmap)
+_gdk_win32_bitmap_to_hrgn (GdkPixmap *pixmap)
 {
   HRGN hRgn = NULL;
   HRGN h;
   DWORD maxRects;
   RGNDATA *pData;
-  GdkImage *image;
+  guchar *bits;
+  gint width, height, bpl;
   guchar *p;
   gint x, y;
 
-  image = GDK_PIXMAP_IMPL_WIN32 (GDK_PIXMAP_OBJECT (pixmap)->impl)->image;
-  g_assert (image->depth == 1);
+  g_assert (GDK_PIXMAP_OBJECT(pixmap)->depth == 1);
+
+  bits = GDK_PIXMAP_IMPL_WIN32 (GDK_PIXMAP_OBJECT (pixmap)->impl)->bits;
+  width = GDK_PIXMAP_IMPL_WIN32 (GDK_PIXMAP_OBJECT (pixmap)->impl)->width;
+  height = GDK_PIXMAP_IMPL_WIN32 (GDK_PIXMAP_OBJECT (pixmap)->impl)->height;
+  bpl = ((width - 1)/32 + 1)*4;
 
   /* For better performances, we will use the ExtCreateRegion()
    * function to create the region. This function take a RGNDATA
@@ -1227,15 +1131,15 @@ _gdk_win32_bitmap_to_region (GdkPixmap *pixmap)
   pData->rdh.nCount = pData->rdh.nRgnSize = 0;
   SetRect (&pData->rdh.rcBound, MAXLONG, MAXLONG, 0, 0);
 
-  for (y = 0; y < image->height; y++)
+  for (y = 0; y < height; y++)
     {
       /* Scan each bitmap row from left to right*/
-      p = (guchar *) image->mem + y * image->bpl;
-      for (x = 0; x < image->width; x++)
+      p = (guchar *) bits + y * bpl;
+      for (x = 0; x < width; x++)
 	{
 	  /* Search for a continuous range of "non transparent pixels"*/
 	  gint x0 = x;
-	  while (x < image->width)
+	  while (x < width)
 	    {
 	      if ((((p[x/8])>>(7-(x%8)))&1) == 0)
 		/* This pixel is "transparent"*/
@@ -1306,100 +1210,52 @@ _gdk_win32_bitmap_to_region (GdkPixmap *pixmap)
   return hRgn;
 }
 
-#ifdef G_ENABLE_DEBUG
-
-gchar *
-gdk_win32_cap_style_to_string (GdkCapStyle cap_style)
+HRGN
+_gdk_win32_gdkregion_to_hrgn (GdkRegion *region,
+			      gint       x_origin,
+			      gint       y_origin)
 {
-  switch (cap_style)
-    {
-#define CASE(x) case x: return #x + strlen ("GDK_CAP_")
-    CASE (GDK_CAP_NOT_LAST);
-    CASE (GDK_CAP_BUTT);
-    CASE (GDK_CAP_ROUND);
-    CASE (GDK_CAP_PROJECTING);
-#undef CASE
-    default: return ("illegal GdkCapStyle value");
-    }
-  /* NOTREACHED */
-  return NULL;
-}
+  HRGN hrgn;
+  RGNDATA *rgndata;
+  RECT *rect;
+  GdkRegionBox *boxes = region->rects;
+  guint nbytes =
+    sizeof (RGNDATAHEADER) + (sizeof (RECT) * region->numRects);
+  int i;
 
-gchar *
-gdk_win32_fill_style_to_string (GdkFill fill)
-{
-  switch (fill)
-    {
-#define CASE(x) case x: return #x + strlen ("GDK_")
-    CASE (GDK_SOLID);
-    CASE (GDK_TILED);
-    CASE (GDK_STIPPLED);
-    CASE (GDK_OPAQUE_STIPPLED);
-#undef CASE
-    default: return ("illegal GdkFill value");
-    }
-  /* NOTREACHED */
-  return NULL;
-}
+  rgndata = g_malloc (nbytes);
+  rgndata->rdh.dwSize = sizeof (RGNDATAHEADER);
+  rgndata->rdh.iType = RDH_RECTANGLES;
+  rgndata->rdh.nCount = rgndata->rdh.nRgnSize = 0;
+  SetRect (&rgndata->rdh.rcBound,
+	   G_MAXSHORT, G_MAXSHORT, G_MINSHORT, G_MINSHORT);
 
-gchar *
-gdk_win32_function_to_string (GdkFunction function)
-{
-  switch (function)
+  for (i = 0; i < region->numRects; i++)
     {
-#define CASE(x) case x: return #x + strlen ("GDK_")
-    CASE (GDK_COPY);
-    CASE (GDK_INVERT);
-    CASE (GDK_XOR);
-    CASE (GDK_CLEAR);
-    CASE (GDK_AND);
-    CASE (GDK_AND_REVERSE);
-    CASE (GDK_AND_INVERT);
-    CASE (GDK_NOOP);
-    CASE (GDK_OR);
-    CASE (GDK_EQUIV);
-    CASE (GDK_OR_REVERSE);
-    CASE (GDK_COPY_INVERT);
-    CASE (GDK_OR_INVERT);
-    CASE (GDK_NAND);
-    CASE (GDK_SET);
-#undef CASE
-    default: return ("illegal GdkFunction value");
-    }
-  /* NOTREACHED */
-  return NULL; 
-}
+      rect = ((RECT *) rgndata->Buffer) + rgndata->rdh.nCount++;
 
-gchar *
-gdk_win32_join_style_to_string (GdkJoinStyle join_style)
-{
-  switch (join_style)
-    {
-#define CASE(x) case x: return #x + strlen ("GDK_JOIN_")
-    CASE (GDK_JOIN_MITER);
-    CASE (GDK_JOIN_ROUND);
-    CASE (GDK_JOIN_BEVEL);
-#undef CASE
-    default: return ("illegal GdkJoinStyle value");
-    }
-  /* NOTREACHED */
-  return NULL; 
-}
+      rect->left = CLAMP (boxes[i].x1 + x_origin,
+			  G_MINSHORT, G_MAXSHORT);
+      rect->right = CLAMP (boxes[i].x2 + x_origin,
+			   G_MINSHORT, G_MAXSHORT);
+      rect->top = CLAMP (boxes[i].y1 + y_origin,
+			 G_MINSHORT, G_MAXSHORT);
+      rect->bottom = CLAMP (boxes[i].y2 + y_origin,
+			    G_MINSHORT, G_MAXSHORT);
 
-gchar *
-gdk_win32_line_style_to_string (GdkLineStyle line_style)
-{
-  switch (line_style)
-    {
-#define CASE(x) case x: return #x + strlen ("GDK_LINE_")
-    CASE(GDK_LINE_SOLID);
-    CASE(GDK_LINE_ON_OFF_DASH);  
-    CASE(GDK_LINE_DOUBLE_DASH);  
-#undef CASE
-    default: return ("illegal GdkLineStyle value");
+      if (rect->left < rgndata->rdh.rcBound.left)
+	rgndata->rdh.rcBound.left = rect->left;
+      if (rect->right > rgndata->rdh.rcBound.right)
+	rgndata->rdh.rcBound.right = rect->right;
+      if (rect->top < rgndata->rdh.rcBound.top)
+	rgndata->rdh.rcBound.top = rect->top;
+      if (rect->bottom > rgndata->rdh.rcBound.bottom)
+	rgndata->rdh.rcBound.bottom = rect->bottom;
     }
-  /* NOTREACHED */
-  return NULL; 
-}
+  if ((hrgn = ExtCreateRegion (NULL, nbytes, rgndata)) == NULL)
+    WIN32_API_FAILED ("ExtCreateRegion");
 
-#endif
+  g_free (rgndata);
+
+  return (hrgn);
+}
