@@ -294,7 +294,15 @@ gtk_cell_view_style_set (GtkWidget *widget,
 static void
 gtk_cell_view_finalize (GObject *object)
 {
-  gtk_cell_view_cell_layout_clear (GTK_CELL_LAYOUT (object));
+  GtkCellView *cellview = GTK_CELL_VIEW (object);
+
+  gtk_cell_view_cell_layout_clear (GTK_CELL_LAYOUT (cellview));
+
+  if (cellview->priv->model)
+     g_object_unref (cellview->priv->model);
+
+  if (cellview->priv->displayed_row)
+     gtk_tree_row_reference_free (cellview->priv->displayed_row);
 
   if (G_OBJECT_CLASS (parent_class)->finalize)
     (* G_OBJECT_CLASS (parent_class)->finalize) (object);
@@ -408,6 +416,7 @@ gtk_cell_view_expose (GtkWidget      *widget,
   GList *i;
   GtkCellView *cellview;
   GdkRectangle area;
+  GtkCellRendererState state;
   gboolean rtl = (gtk_widget_get_direction(widget) == GTK_TEXT_DIR_RTL);
 
   cellview = GTK_CELL_VIEW (widget);
@@ -437,9 +446,11 @@ gtk_cell_view_expose (GtkWidget      *widget,
       g_object_unref (G_OBJECT (gc));
     }
 
-  /* set cell data (if applicable) */
+  /* set cell data (if available) */
   if (cellview->priv->displayed_row)
     gtk_cell_view_set_cell_data (cellview);
+  else if (cellview->priv->model)
+    return FALSE;
 
   /* render cells */
   area = widget->allocation;
@@ -448,6 +459,11 @@ gtk_cell_view_expose (GtkWidget      *widget,
   area.x = widget->allocation.x + (rtl ? widget->allocation.width : 0); 
   area.y = widget->allocation.y;
 
+  if (GTK_WIDGET_STATE (widget) == GTK_STATE_PRELIGHT)
+    state = GTK_CELL_RENDERER_PRELIT;
+  else
+    state = 0;
+      
   /* PACK_START */
   for (i = cellview->priv->cell_list; i; i = i->next)
     {
@@ -467,7 +483,7 @@ gtk_cell_view_expose (GtkWidget      *widget,
                                 event->window,
                                 widget,
                                 /* FIXME! */
-                                &area, &area, &event->area, 0);
+                                &area, &area, &event->area, state);
 
       if (!rtl)                                           
          area.x += info->real_width;
@@ -494,7 +510,7 @@ gtk_cell_view_expose (GtkWidget      *widget,
                                 widget->window,
                                 widget,
                                 /* FIXME ! */
-                                &area, &area, &event->area, 0);
+                                &area, &area, &event->area, state);
       if (rtl)
          area.x += info->real_width;
     }
@@ -537,15 +553,7 @@ gtk_cell_view_set_cell_data (GtkCellView *cellview)
       GSList *j;
       GtkCellViewCellInfo *info = i->data;
 
-      if (info->func)
-        {
-          (* info->func) (GTK_CELL_LAYOUT (cellview),
-                          info->cell,
-                          cellview->priv->model,
-                          &iter,
-                          info->func_data);
-          continue;
-        }
+      g_object_freeze_notify (G_OBJECT (info->cell));
 
       for (j = info->attributes; j && j->next; j = j->next->next)
         {
@@ -559,6 +567,15 @@ gtk_cell_view_set_cell_data (GtkCellView *cellview)
                                  property, &value);
           g_value_unset (&value);
         }
+
+      if (info->func)
+	(* info->func) (GTK_CELL_LAYOUT (cellview),
+			info->cell,
+			cellview->priv->model,
+			&iter,
+			info->func_data);
+
+      g_object_thaw_notify (G_OBJECT (info->cell));
     }
 }
 
@@ -631,22 +648,20 @@ gtk_cell_view_cell_layout_add_attribute (GtkCellLayout   *layout,
 static void
 gtk_cell_view_cell_layout_clear (GtkCellLayout *layout)
 {
-  GList *i;
   GtkCellView *cellview = GTK_CELL_VIEW (layout);
 
   g_return_if_fail (GTK_IS_CELL_VIEW (cellview));
 
-  for (i = cellview->priv->cell_list; i; i = i->next)
+  while (cellview->priv->cell_list)
     {
-      GtkCellViewCellInfo *info = (GtkCellViewCellInfo *)i->data;
+      GtkCellViewCellInfo *info = (GtkCellViewCellInfo *)cellview->priv->cell_list->data;
 
       gtk_cell_view_cell_layout_clear_attributes (layout, info->cell);
       g_object_unref (G_OBJECT (info->cell));
       g_free (info);
+      cellview->priv->cell_list = g_list_delete_link (cellview->priv->cell_list, 
+						      cellview->priv->cell_list);
     }
-
-  g_list_free (cellview->priv->cell_list);
-  cellview->priv->cell_list = NULL;
 }
 
 static void
